import pickle
import os
import math
import pygame
import pygame.font
import pygame.image
from story import *
from resource import absolute_path as abspath
from scenedata import scenes

import renderer

from pygame import Rect # TODO: check if this is still needed here
# once the renderer is factorised out correctly, this import should
# become obsolete (?)

__DEBUG__ = False

class Game:
    def __init__(self, story_file, debug=False, begin_sentence=None,
      fullscreen=False, android=None, newgame=False, newGraphics=False):
        global __DEBUG__
        __DEBUG__ = debug
        
        screen = False
        if (debug and newGraphics):
            try:
                screen = glrenderer.gl_supported()
            except NameError:
                pass # fall back on renderer module
        if screen:
            print 'new graphics!'
            self.renderer = glrenderer.Renderer(size=(800, 480),
              windowTitle='Gregor Samsa', fullscreenMode=fullscreen,
              debugMode=debug, screen=screen)
        else:
            self.renderer = renderer.Renderer(size=(800, 480),
              windowTitle='Gregor Samsa', fullscreenMode=fullscreen,
              debugMode=debug)
        
        self.exit = False
        
        self.renderer.set_scene(scenes["default"])

        if not os.path.isfile("visited.dat") or newgame == True:
            self.story = Story(abspath(story_file))
        else:
            visitedfile = open("visited.dat", "r")
            self.story = pickle.load(visitedfile)
            visitedfile.close()
        self.sentence_words = []
        self.new_sentence_words = []
        self.update_all = True
        self.drag_word = None
        """ android is needed on android phones """
        self.android = android 
        if begin_sentence is None:
            self.new_sentence('0')
        else:
            self.new_sentence(begin_sentence)


    def load_sentence(self, sentence):
        self.sentence = sentence
        sentence.visited = True
        self.sentence_words = [DisplayWord(w, self) for w in sentence]
        self.new_sentence_words = [
          DisplayWord(w, self) for w in sentence.get_newwords()]
        self.all_words = self.sentence_words + self.new_sentence_words
        if sentence.image:
            self.renderer.set_image(sentence.image)
    
    def new_sentence(self, pid):
        self.load_sentence(self.story.get_object(pid))
        if self.sentence.scene:
            if scenes.has_key(self.sentence.scene):
                self.renderer.set_scene(scenes[self.sentence.scene])
            else:
                self.renderer.set_scene(scenes["default"])
        self.renderer.layout_sentence(sentenceWords=self.sentence_words)
        self.renderer.layout_new_words(sentenceWords=self.sentence_words,
          newSentenceWords=self.new_sentence_words)
        self.update_all = True
        scene = int(pid.split('.')[0])

    def process_events(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.exit = True
            elif event.type == pygame.KEYUP:
                if event.key == pygame.K_ESCAPE or event.key == pygame.K_q:
                    self.exit = True
            elif event.type == pygame.MOUSEMOTION:
                self.update_all = True
                if self.drag_word:
                    self.drag_word.update_drag(event.pos)
                    for word in self.all_words:
                        if word == self.drag_word:
                            continue
                        word.update_hover_rect(self.drag_word.rect)
                else:
                    for word in self.all_words:
                        word.update_hover(event.pos)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                word = self.get_first_moveable_word_at_coordinates(event.pos)
                if not word:
                    continue
                if (word.word.has_action("move")
                  or word.word.has_action("remove")):
                    try:
                        word.original_rect = word.rect.copy()
                    except AttributeError:
                        import copy
                        word.original_rect = copy.deepcopy(word.rect)
                    word.follow_mouse(event.pos)
                    self.drag_word = word
            elif event.type == pygame.MOUSEBUTTONUP:
                if not self.drag_word:
                    continue
                word = self.drag_word
                self.drag_word = None
                word.mouse_offset=None
                if word.word.has_action("remove"):
                    #check if the word was dragged far enough
                    if not word.rect.colliderect(word.original_rect):
                        self.new_sentence(word.word.do_action("remove"))
                    else:
                        word.set_pos(word.original_rect.topleft)
                        self.update_all = True
                word.check_collisions(self.new_sentence_words+self.sentence_words)    
    
    
    def get_first_moveable_word_at_coordinates(self, pos):
        for word in self.new_sentence_words + self.sentence_words:
            if word.word.is_actionable() and word.rect.collidepoint(pos):
                return word
        return None

    def run(self):
        while not self.exit:
            if self.android:
                if self.android.check_pause():
                    android.wait_for_resume()
            self.process_events()
            self.renderer.draw(allWords=self.all_words, story=self.story, update_all=self.update_all)
            self.update_all = False
        visitedfile = open("visited.dat", "w")
        pickle.dump(self.story, visitedfile)
        visitedfile.close()

class DisplayWord:
    def __init__(self, word, game):
        self.word = word
        self.game = game

        #prerender words
        text = word.word
        self.surface = game.renderer.render_text(text,
          game.renderer.font_color).convert_alpha()
        if word.is_actionable():
            self.active_surface = game.renderer.render_text(
              text, game.renderer.font_active_color).convert_alpha()
        else:
            self.active_surface = self.surface

        #init UI parameters
        self.is_hover = False
        self.rect = Rect(0, 0, self.surface.get_width(),
          self.surface.get_height())
        self.mouse_offset=None
        self.original_rect = Rect((0, 0), (0, 0))

    def set_pos(self, pos):
        self.rect.topleft = pos

    def width(self):
        return self.rect.width

    def height(self):
        return self.rect.height
    
    def check_collisions(self, words):
        for word in words:
           if self.rect.colliderect(word.rect) and (word != self):
              if word.word.has_action("move"):
                  word.evade_rect(self.rect)
              if word.word.has_action("replace"):
                  r = word.word.do_action("replace", new=self.word)
                  if r:
                    self.game.new_sentence(r)
   
    def evade_rect(self, rect):
        if not self.word.has_action("move"):
            return None
        vector = (self.rect.topleft[0] - rect.topleft[0],
          self.rect.topleft[1] - rect.topleft[1])
        vl = math.sqrt(vector[0] ** 2 + vector[1] **2 )
        if vl == 0:
            vl = 1
        vector = (vector[0] / vl * 20, vector[1] / vl * 20)
        self.set_pos((self.rect.topleft[0] + vector[0],
          self.rect.topleft[1] + vector[1]))
        self.game.renderer.check_page(self.rect)
      

    def update_hover(self, mouse_pos):
        self.is_hover = self.rect.collidepoint(mouse_pos)

    def update_hover_rect(self, rect):
        self.is_hover = self.rect.colliderect(rect)

    def update_drag(self, mouse_pos):
        if self.mouse_offset:
          self.set_pos((mouse_pos[0]-self.mouse_offset[0],mouse_pos[1]-self.mouse_offset[1]))
          self.game.renderer.check_page(self.rect)

    def follow_mouse(self,mouse_pos):
        self.mouse_offset = (mouse_pos[0] - self.rect.topleft[0],
          mouse_pos[1] - self.rect.topleft[1])
    
    def draw(self, screen):

        if self.is_hover or self.mouse_offset or __DEBUG__ or self.game.android:
            surface = self.active_surface
        else:
            surface = self.surface
        screen.blit(surface, self.rect.topleft)
        


